#include <windows.h>
#include <io.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <time.h>

#define NUMWORDS 9

#define NL "  \n"

HANDLE htimeout;

// Taken from https://docs.microsoft.com/en-US/windows/console/clearing-the-screen

CONSOLE_SCREEN_BUFFER_INFO csbi; // Global, so it can be compared against

void cls(HANDLE hConsole)
{
  COORD coordScreen = { 0, 0 };    // home for the cursor
  DWORD cCharsWritten;
  DWORD dwConSize;

  // Get the number of character cells in the current buffer.
  if (!GetConsoleScreenBufferInfo(hConsole, &csbi))
    {
      return;
    }

  dwConSize = csbi.dwSize.X * csbi.dwSize.Y;

  // Fill the entire screen with blanks.
  if (!FillConsoleOutputCharacter(hConsole,        // Handle to console screen buffer
				  (TCHAR)' ',      // Character to write to the buffer
				  dwConSize,       // Number of cells to write
				  coordScreen,     // Coordinates of first cell
				  &cCharsWritten)) // Receive number of characters written
    {
      return;
    }

  // Get the current text attribute.
  if (!GetConsoleScreenBufferInfo(hConsole, &csbi))
    {
      return;
    }

  // Set the buffer's attributes accordingly.
  if (!FillConsoleOutputAttribute(hConsole,         // Handle to console screen buffer
				  csbi.wAttributes, // Character attributes to use
				  dwConSize,        // Number of cells to set attribute
				  coordScreen,      // Coordinates of first cell
				  &cCharsWritten))  // Receive number of characters written
    {
      return;
    }

  // Put the cursor at its home coordinates.
  SetConsoleCursorPosition(hConsole, coordScreen);
}

// A separate thread waits for the timeout event rather than using a timer
// APC routine, because the latter requires the thread to be alertable, and
// that's a bit difficult to ensure, in particular with regards to the
// unexpected situation that would trigger the timeout to begin with.

DWORD WINAPI exit_on_timeout_thread(LPVOID arg) {
  // Any reason for WaitForSingleObject() to return is good enough to quit.
  WaitForSingleObject(htimeout, INFINITE);

  fprintf(stderr, "\nTimed out on receiving data.\n");

  _exit(1); // Plain exit() won't always kill main thread.

  return 0; // Never reached.
}

static int allread(int fd, unsigned char *buf, int len) {
  int done = 0;
  int rc;

  while (done < len) {
    rc = _read(fd, buf + done, len - done);

    if ((rc < 0) && (errno == EINTR))
      continue;

    if (rc < 0) {
      perror("allread() failed to read");
      exit(1);
    }

    if (rc == 0)
      return done; // EOF. It's fine

    done += rc;
  }

  return done;
}

int main(int argc, char *argv[]) {
  HANDLE hout = GetStdHandle(STD_OUTPUT_HANDLE);
  CONSOLE_SCREEN_BUFFER_INFO csbi_now;
  int index = 0;
  char fname[40];
  const char twiddle[] = { '|', '/', '-', '\\' };
  int twidpos = 0;
  int firstloop = 1;
  int i;
  LARGE_INTEGER timeout_time;

  int32_t words[NUMWORDS], firstwords[NUMWORDS];

  if ((argc > 2) ||
      ((argc == 2) && (sscanf(argv[1], "%d", &index) != 1))) {
    fprintf(stderr, "Usage: %s [numeric-index]\n", argv[0]);
    exit(1);
  }

  timeout_time.QuadPart = -30000000LL; // 3000 ms
  htimeout = CreateWaitableTimer(NULL, TRUE, NULL);

  if (!htimeout) {
    fprintf(stderr, "Failed to create timer!\n");
    exit(1);
  }

  CreateThread(NULL, 0, exit_on_timeout_thread, NULL, 0, NULL);

  if (_snprintf(fname, sizeof(fname),
		"\\\\.\\xillyusb_%02d_diagnostics", index) >= sizeof(fname)) {
    fprintf(stderr, "The device file's name turned out too long!\n");
    exit(1);
  }

  for (i=0; i<NUMWORDS; i++)
    firstwords[i] = 0; // Just to silence warning

  cls(hout);

  while (1) {
    int fd, rc;
    time_t ltime;
    int32_t power_ctrl, u2_timeout;
    COORD coordScreen = { 0, 0 };

    if (!SetWaitableTimer(htimeout, &timeout_time, 0, NULL, NULL, 0)) {
      fprintf(stderr, "Failed to set timer!\n");
      exit(1);
    }

    fd = _open(fname, O_RDONLY | _O_BINARY);

    if (fd < 0) {
      perror("open");
      exit(1);
    }

    rc = allread(fd, (void *) words, sizeof(words));
    _close(fd);

    if (rc < sizeof(words)) {
      fprintf(stderr, "Read too little data: Expected %d bytes (at least) and got %d.\n",
	      (int) sizeof(words), rc);
      exit(1);
    }

    if (words[0] != 1) {
      fprintf(stderr, "Unexpected diagnostics data version %d, should be 1\n", words[0]);
      exit(1);
    }

    if (firstloop) {
      firstloop = 0;

      for (i=0; i<NUMWORDS; i++)
	firstwords[i] = words[i];
    }

    power_ctrl = words[8];

    for (i=0; i<NUMWORDS; i++)
      words[i] -= firstwords[i];

    // If the console has the same dimensions as when it was cleared last,
    // just move the cursor to home, so flickering is avoided. Otherwise
    // clear the screen, or the text might become garbled.

    if (GetConsoleScreenBufferInfo(hout, &csbi_now) &&
	(csbi_now.dwSize.X == csbi.dwSize.X) &&
	(csbi_now.dwSize.Y == csbi.dwSize.Y) &&
	(csbi_now.srWindow.Left == csbi.srWindow.Left) &&
	(csbi_now.srWindow.Right == csbi.srWindow.Right) &&
	(csbi_now.srWindow.Top == csbi.srWindow.Top) &&
	(csbi_now.srWindow.Bottom == csbi.srWindow.Bottom))
      SetConsoleCursorPosition(hout, coordScreen); // Move cursor to home
    else
      cls(hout);

    // asctime() ends with a "\n" before the terminating null string. This
    // may cause messy appearance if the timestamp gets shorter after a while
    // (which should probably never happen).
    ltime = time(NULL);
    printf("XillyUSB event counts %c %s" NL,
	   twiddle[twidpos], asctime(localtime(&ltime)));

    twidpos = (twidpos + 1) & 3;

    printf("Errors on FPGA to host link:" NL
	   "===========================" NL
	   "Bad packets received by link partner: %d" NL
	   "Recovery requests by link partner: %d" NL NL
	   "Errors on host to FPGA link:" NL
	   "===========================" NL
	   "Bad packet received by device: %d" NL
	   "Errors detected during link idle: %d" NL
	   "Recovery requests by device: %d" NL NL,
	   words[2], words[1], words[4], words[5], words[3]);

   printf("Power management:" NL
	  "================" NL
	  "XillyUSB device's power policy:" NL
	  "%s" NL NL,
	  (power_ctrl & 0x20000) ? "AGREE to low power state" :
	  (power_ctrl & 0x10000) ? "REFUSE to low power state" :
	  "Balanced (agree to low power state when no device file is open)");

   u2_timeout = power_ctrl & 0xff;

   printf("PORT_U2_TIMEOUT: ");

   if ((u2_timeout == 0) || (u2_timeout == 0xff))
     printf("(disabled)" NL NL);
   else
     printf("%.3f ms" NL NL, 0.256 * (double) u2_timeout);

   printf("Low power transitions made: %d" NL
	  "Low power transitions refused by device: %d" NL,
	  words[6], words[7]);

    Sleep(250);
  }
}
